#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>

#include "ximea_cam_pcie.h"

#define BUF_PAGES atoi(argv[2])
#define BUF_LENGTH (4096*BUF_PAGES)
#define BUF_NUMBER atoi(argv[1])
#define DESC_PAGES (BUF_PAGES*16/4080+1)
#define DESC_LENGTH (4096*DESC_PAGES)


int main(int argc, char** argv)
{
	int fd = open("/dev/ximea00", O_RDWR);
	if(fd < 0) return 1;
	printf("Opened /dev/ximea00\n");
	void* buf[BUF_NUMBER];
	void* hdr[BUF_NUMBER];
	void* desc[BUF_NUMBER];
	for(int i = 0; i < BUF_NUMBER; i++)
	{
       		buf[i] = mmap(0, BUF_LENGTH, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, 0, 0);
		if(!buf[i]) return 2;
		madvise(buf[i], BUF_LENGTH, MADV_DONTFORK);
       		hdr[i] = mmap(0, 4096, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, 0, 0);
		if(!hdr[i]) return 2;
		madvise(hdr[i], 4096, MADV_DONTFORK);
       		desc[i] = mmap(0, DESC_LENGTH, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, 0, 0);
		if(!desc[i]) return 2;
		madvise(desc[i], DESC_LENGTH, MADV_DONTFORK);
	}
	printf("Allocated %dx%d pages\n", BUF_NUMBER, BUF_PAGES);
	ximea_cam_pcie_lockmem_args* lock = malloc(sizeof(*lock) * BUF_NUMBER * 3);
	ximea_cam_pcie_unlockmem_args* unlock = malloc(sizeof(*unlock) * (BUF_NUMBER * 3 + BUF_PAGES));
	for(int i = 0; i < BUF_NUMBER; i++)
	{
		lock[i*3].buffer = buf[i];
		lock[i*3].bytes = BUF_LENGTH;
		lock[i*3].pages = BUF_PAGES;
		lock[i*3].out = (void*)&unlock[i*3];
		int ret = ioctl(fd, XIMEA_CAM_PCIE_IOCLOCKMEM, &lock[i*3]);
		if(ret) return ret;
		lock[i*3+1].buffer = hdr[i];
		lock[i*3+1].bytes = 4096;
		lock[i*3+1].pages = 1;
		lock[i*3+1].out = (void*)&unlock[i*3+1];
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCLOCKMEM, &lock[i*3+1]);
		if(ret) return ret;
		lock[i*3+2].buffer = desc[i];
		lock[i*3+2].bytes = DESC_LENGTH;
		lock[i*3+2].pages = DESC_PAGES;
		lock[i*3+2].out = (void*)&unlock[i*3+2];
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCLOCKMEM, &lock[i*3+2]);
		if(ret) return ret;
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCSYNCMEM4DEV, &unlock[i*3+2]);
		if(ret) return ret;
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCSYNCMEM4DEV, &unlock[i*3+1]);
		if(ret) return ret;
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCSYNCMEM4DEV, &unlock[i*3]);
		if(ret) return ret;
	}
	for(int i = 0; i < BUF_NUMBER; i++)
	{
		int ret = ioctl(fd, XIMEA_CAM_PCIE_IOCSYNCMEM4CPU, &unlock[i*3+1]);
		if(ret) return ret;
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCSYNCMEM4CPU, &unlock[i*3]);
		if(ret) return ret;
	}
	printf("Locked pages\n");
	for(int i = 0; i < BUF_NUMBER; i++)
	{
		int ret = ioctl(fd, XIMEA_CAM_PCIE_IOCUNLOCKMEM, &unlock[i*3+1]);
		if(ret) return ret;
		munmap(hdr[i], 4096);
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCUNLOCKMEM, &unlock[i*3+2]);
		if(ret) return ret;
		munmap(desc[i], DESC_LENGTH);
		ret = ioctl(fd, XIMEA_CAM_PCIE_IOCUNLOCKMEM, &unlock[i*3]);
		if(ret) return ret;
	}
	printf("Unlocked pages\n");
	free(unlock);
	free(lock);
	for(int i = 0; i < BUF_NUMBER; i++)
	{
		munmap(buf[i], BUF_LENGTH);
	}
	printf("Freed memory\n");
	close(fd);
	printf("Closed device\n");
}
